"""
Module for command line interface features
"""

import argparse

DESCRIPTION = """A Python tool for generating static websites based on object models"""


def get_parser_options():
    """ Generate CLI options """
    parser = argparse.ArgumentParser(
        prog='webstaterator',
        description=DESCRIPTION
    )

    arg_website_short = '-w'
    arg_website_long = '--website'
    arg_website_desc = 'Website decription file'

    arg_output_short = '-o'
    arg_output_long = '--output'
    arg_output_desc = 'Output path'

    action_parser = parser.add_subparsers(
        title='Actions',
        description='Webstaterator actions',
        dest='action'
    )
    action_validate = action_parser.add_parser(
        'validate',
        help='Validates the provided website description file'
    )
    action_validate.add_argument(
        arg_website_short,
        arg_website_long,
        help=arg_website_desc,
        required = True
    )

    action_build = action_parser.add_parser(
        'build',
        help="""Builds a website in the given folder based
on the provided website description file"""
        )
    action_build.add_argument(
        arg_website_short,
        arg_website_long,
        help=arg_website_desc,
        required = True,
    )
    action_build.add_argument(
        arg_output_short,
        arg_output_long,
        help=arg_output_desc,
        required = True,
    )

    action_template = action_parser.add_parser(
        'template',
        help='Generates a blank website description file'
    )

    action_template.add_argument(
        arg_output_short,
        arg_output_long,
        help=arg_output_desc,
        required = True,
    )

    return parser



""" pytest module to hold tests for webstaterator/website.py """

import os
import json
import shutil

import pytest

from webstaterator.website import Website
from webstaterator.website import InvalidPagesException
from webstaterator.website import InvalidModelException
from webstaterator.website import InvalidWebsiteException
from webstaterator.website import PageAttributes

def test_create_website_instance():
    """ Check an instance of website can be created """
    assert isinstance(Website(),Website)

#--- test data generators ------------------------------------------------------
def create_page(name,template,filename):
    """ helper method to create a valid page """
    return {
            'name':name,
            'template':template,
            'filename':filename
        }

def create_multipage(name,template,model_list,model_id,filename):
    """ helper method to create a valid page """
    return {
            'name':name,
            'template':template,
            'list':model_list,
            'id':model_id,
            'filename':filename
        }

def get_valid_pages():
    """ helper method to create a valid list of pages """
    return [
            create_page('test1','index.html','index.html'),
            create_page('test2','links.html','items.html')
    ]

def get_duplicate_pages():
    """ helper method to create a list of pages containing duplicates """
    return [
            create_page('test1','index.html','index.html'),
            create_page('test2','links.html','items.html'),
            create_page('test1','index.html','index.html'),
    ]

def get_invalid_pages():
    """ helper method to create an invalid list of pages """
    return [
            create_page('test1','index.html','index.html'),
            {
                'name':'invalid',
                'template':'error.html'
            }
    ]

def get_valid_pages_with_lists():
    """ helper method that creates a list of pages including one multipage """

    return [
            create_page('test1','index.html','index.html'),
            create_page('test2','links.html','items.html'),
            create_multipage('test3','items.html','items','id','item-{}.html')
    ]

def get_basic_website(pages):
    return {
        'name':'test',
        'template':'templates',
        'pages':pages
    }

def get_website_with_model(pages,model):
    website = get_basic_website(pages)
    website['model'] = model
    return website

@pytest.fixture()
def get_test_website_model():
    website_dict = get_basic_website(get_valid_pages())
    website_path = os.path.join(os.getcwd(),'test_website.js')

    with open(website_path,'w') as website_stream:
        json.dump(website_dict,website_stream)

    yield website_dict, website_path

    os.remove(website_path)

@pytest.fixture()
def get_test_asset_folder():
    asset_path = os.path.join(os.getcwd(),"tests_assets")
    if(os.path.exists(asset_path)):
        shutil.rmtree(asset_path)

    os.mkdir(asset_path)
    yield asset_path
    shutil.rmtree(asset_path)

#--- load tests ----------------------------------------------------------------
def test_load_with_dict():
    test_website = Website()
    website_dict = get_basic_website(get_valid_pages())
    test_website.load(website_dict)

    assert isinstance(test_website,Website)
    assert test_website.name == website_dict['name']
    assert test_website.template_path == website_dict['template']
    assert test_website.pages == website_dict['pages']

def test_load_dict_with_bad_template():
    test_website = Website()

    website_with_bad_template_path = get_basic_website(get_valid_pages())
    website_with_bad_template_path['template'] = 'missing_folder'

    with pytest.raises(InvalidWebsiteException):
        test_website.load(website_with_bad_template_path)

def test_load_with_json_file(get_test_website_model):
    test_website = Website()
    website_dict = get_test_website_model[0]
    test_website.load(get_test_website_model[1])

    assert isinstance(test_website,Website)
    assert test_website.name == website_dict['name']
    assert test_website.template_path == website_dict['template']
    assert test_website.pages == website_dict['pages']

def test_load_with_json_string():
    test_website = Website()
    website_dict = get_basic_website(get_valid_pages())
    json_website = json.dumps(website_dict)
    test_website.load(json_website)

    assert isinstance(test_website,Website)
    assert test_website.name == website_dict['name']
    assert test_website.template_path == website_dict['template']
    assert test_website.pages == website_dict['pages']

def test_load_with_assets(get_test_asset_folder):
    test_website = Website()
    website_dict = get_basic_website(get_valid_pages())
    website_dict['assets'] = [get_test_asset_folder]
    test_website.load(website_dict)

    assert test_website.asset_paths == website_dict['assets']

def test_to_dict():
    test_website = Website()
    website_dict = get_basic_website(get_valid_pages())
    test_website.load(website_dict)

    website_obj = test_website.to_dict()

    assert website_obj['name'] == website_dict['name']
    assert website_obj['template'] == website_dict['template']
    assert website_obj['pages'] == website_dict['pages']

#--- model tests ---------------------------------------------------------------
def test_set_model_with_dict():
    """ Test that a model can be set as a dict """
    test_website = Website()
    assert test_website.model == {}

    test_model = {'a':1,'b':2,'c':3}
    test_website.model = test_model
    assert test_website.model == test_model

def test_set_model_with_json_string():
    """ test that a model can be set using a json string """
    test_website = Website()
    assert test_website.model == {}

    test_model = """{"a":1,"b":2,"c":3}"""
    test_website.model = test_model
    json_model = json.loads(test_model)
    assert test_website.model == json_model

def test_set_model_with_json_file():
    """ test that a model can be set using a json file """
    test_website = Website()
    assert test_website.model == {}

    test_model = """{"a":1,"b":2,"c":3}"""
    file_path = os.path.join(os.getcwd(),"test_model.json")
    with open(file_path,'w') as file_out:
        file_out.write(test_model)
    test_website.model = file_path
    os.remove(file_path)
    json_model = json.loads(test_model)

    assert test_website.model == json_model

#--- pages tests ---------------------------------------------------------------
def test_pages_validation_empty_list():
    """ test to check that setting pages to an empty list results in a raised
    exception """
    test_website = Website()
    with pytest.raises(InvalidPagesException):
        test_website.pages = []

def test_valid_pages():
    """ test to check that pages accepts a valid list of pages """
    test_website = Website()
    test_website.pages =get_valid_pages()
    assert test_website.pages == get_valid_pages()

def test_invalid_pages():
    """ test to check that setting pages to an invalid list of pages raises
    an exception """
    test_website = Website()
    with pytest.raises(InvalidPagesException):
        test_website.pages = get_invalid_pages()

def test_duplicate_pages():
    test_website = Website()
    with pytest.raises(InvalidPagesException):
        test_website.pages = get_duplicate_pages()

#--- link tests ----------------------------------------------------------------
def test_validate_list_page_with_valid_filename():
    page = create_multipage(
        'test',
        'tmp.html',
        'model',
        None,
        'out-{}.html'
    )
    assert Website.validate_list_page(page) == True


def test_validate_list_page_with_invalid_filename():
    page = create_multipage(
        'test',
        'tmp.html',
        'model',
        None,
        'out.html'
    )
    with pytest.raises(InvalidPagesException):
        Website.validate_list_page(page)


def test_get_target_model_with_valid_no_dot_target():
    target_obj = "test"
    obj = {
        "test":"value",
        "another":"value"
    }

    target = Website.get_target_model(target_obj, obj)
    assert target == obj[target_obj]


def test_get_target_model_with_valid_dot_target():
    target_obj = "missing"
    obj = {
        "test":"value",
        "another":"value"
    }

    with pytest.raises(InvalidModelException):
        target = Website.get_target_model(target_obj, obj)


def test_get_target_model_with_invalid_no_dot_target():
    target_obj = "test.nested.value"
    obj = {
        "test":{
            "nested":{
                "value":"hello",
                "value2":"world"
            }
        },
        "another":"value"
    }

    target = Website.get_target_model(target_obj, obj)
    assert target == obj["test"]["nested"]["value"]


@pytest.mark.parametrize('target_obj',[
    'test.nested.missing',
    'test.missing.value',
    'missing.nested.value',
    'test.nested.']
)
def test_get_target_model_with_invalid_dot_target(target_obj):
    obj = {
        "test":{
            "nested":{
                "value":"hello",
                "value2":"world"
            }
        },
        "another":"value"
    }

    with pytest.raises(InvalidModelException):
        target = Website.get_target_model(target_obj, obj)


def test_basic_link_generation():
    """Test single page link generation"""
    test_website = Website()
    pages = get_valid_pages()
    test_website.pages = pages
    test_website.model = """{"a":1,"b":2,"c":3}"""
    assert test_website.generate_links() is True
    expected_links = {p[PageAttributes.NAME]:[p[PageAttributes.FILENAME]] for p in pages}
    assert test_website.links == expected_links

def test_list_link_generation():
    """Test single page link generation"""
    test_website = Website()
    pages = get_valid_pages_with_lists()
    test_website.pages = pages
    test_website.model = """{"a":1,"b":2,"c":3,"items":[{"id":1},{"id":2}]}"""

    assert test_website.generate_links() is True

    expected_links = {
                "test1":["index.html"],
                "test2":["items.html"],
                "test3":["item-1.html","item-2.html"]
    }

    assert test_website.links == expected_links

def test_list_link_generation_with_dict():
    """Test single page link generation"""
    test_website = Website()
    pages = get_valid_pages_with_lists()
    test_website.pages = pages
    test_website.model = """{"a":1,"b":2,"c":3,"items":{"1":{"id":1},"2":{"id":2}}}"""

    assert test_website.generate_links() is True

    expected_links = {
                "test1":["index.html"],
                "test2":["items.html"],
                "test3":["item-1.html","item-2.html"]
    }

    assert test_website.links == expected_links

def test_list_link_target_model():
    """Test single page link generation"""
    test_website = Website()
    pages = get_valid_pages_with_lists()
    test_website.pages = pages
    test_website.model = """{"a":1,"b":2,"c":3,"items":[{"id":1},{"id":2}]}"""

    assert test_website.generate_links() is True

    expected_targets = {
                "index.html":None,
                "items.html":None,
                "item-1.html":{"id":1},
                "item-2.html":{"id":2}
    }

    for url, target in expected_targets.items():
        assert test_website.get_target(url) == target

def test_deep_list_link_generation():
    """Test multipage link generation"""
    test_website = Website()
    pages = get_valid_pages()
    pages.append(create_multipage("test3","item.html","items.foo.bar","id","item-{}.html"))
    test_website.pages = pages
    test_website.model = {
        "a":1,
        "b":2,
        "c":3,
        "items":{
            "foo":{
                "bar":[
                    {"id":"dog"},
                    {"id":"cat"},
                    {"id":"fish"},
                ]
            }
        }
    }

    assert test_website.generate_links() is True

    expected_links = {
                "test1":["index.html"],
                "test2":["items.html"],
                "test3":["item-dog.html","item-cat.html","item-fish.html"]
    }

    assert test_website.links == expected_links

def test_list_link_invalid_filename():
    """Test multipage invalid filename"""
    test_website = Website()
    pages = get_valid_pages()
    pages.append(create_multipage("test3","item.html","items","id","item.html"))
    test_website.pages = pages
    test_website.model = """{"a":1,"b":2,"c":3,"items":[{"id":1},{"id":2}]}"""

    with pytest.raises(InvalidPagesException):
        test_website.generate_links()

def test_list_link_missing_obj():
    """Test multipage model missing object"""
    test_website = Website()
    pages = get_valid_pages()
    pages.append(create_multipage("test3","item.html","items.foo.bar","id","item-{}.html"))
    test_website.pages = pages
    test_website.model = """{"a":1,"b":2,"c":3,"items":[{"id":1},{"id":2}]}"""

    with pytest.raises(InvalidModelException):
        test_website.generate_links()

def test_list_link_missing_id():
    """Test multipage model missing id"""
    test_website = Website()
    pages = get_valid_pages()
    pages.append(create_multipage("test3","item.html","items","foo","item-{}.html"))
    test_website.pages = pages
    test_website.model = """{"a":1,"b":2,"c":3,"items":[{"id":1},{"id":2}]}"""

    with pytest.raises(InvalidModelException):
        test_website.generate_links()

def test_list_link_missing_some_ids():
    """Test multipage model missing some ids"""
    test_website = Website()
    pages = get_valid_pages()
    pages.append(create_multipage("test3","item.html","items","id","item-{}.html"))
    test_website.pages = pages
    test_website.model = """{"a":1,"b":2,"c":3,"items":[{"id":1},{"foo":2}]}"""

    with pytest.raises(InvalidModelException):
        test_website.generate_links()



"""
Webstaterator

A Python tool for generating static websites based on object models

Documentation: https://gitlab.com/Jon.Keatley.Folio/webstaterator
Gitlab: https://gitlab.com/Jon.Keatley.Folio/webstaterator
PyPi: https://pypi.org/project/webstaterator/

Created by Jon Keatley (http://jon-keatley.com)
Named by Sasha Siegel. It is her fault!

Copyright Jon Keatley 2021

"""

import sys
import os
import shutil
import json

from webstaterator.website import Website, PageAttributes
from webstaterator.templateengine import TemplateEngine

class Webstaterator:
    """ A Python tool for generating static websites based on object models """

    @staticmethod
    def validate(website_model):
        """ Validates a websites settings """
        website = Website()

        try:
            website.load(website_model)
            templates = TemplateEngine(website.template_path)

            #check all templates exist
            pages = website.pages
            missing_page_format = "Template {} for page {} was not found"

            has_error = False
            for page in pages:
                if not templates.has_template(page[PageAttributes.TEMPLATE]):
                    err = missing_page_format.format(
                        page[PageAttributes.TEMPLATE],
                        page[PageAttributes.NAME]
                    )
                    has_error = True
                    print(err)


            website.generate_links()

            if not has_error:
                print("Ok")
        except Exception as ex: # pylint: disable=W0703
            print("An error was found with your website!")
            print(f"Error: {type(ex).__name__}. Message: {ex.args}")


    @staticmethod
    def build(website_model, path):
        """ Builds a website based on the provided settings """

        # add check for path
        if os.path.isdir(path):
            shutil.rmtree(path)
        os.mkdir(path)

        print("Loading ...")
        website = Website()
        try:
            website.load(website_model)
            website.generate_links()
            templates = TemplateEngine(website.template_path)
            templates.model = website.model
            templates.links = website.links
        except Exception as ex: # pylint: disable=W0703
            print("An error was found with your website!")
            print(f"Error: {type(ex).__name__}. Message: {ex.args}")
            sys.exit(-1)

        links = website.links

        if website.asset_paths is not None:
            print("copying assets")
            for asset in website.asset_paths:
                if os.path.isdir(asset):
                    dir_name = os.path.basename(asset)
                    new_asset_path = os.path.join(path,dir_name)

                    shutil.copytree(
                        asset,
                        new_asset_path
                    )
                else:
                    print(f"Asset path {website.asset_paths} does not exist")
                    sys.exit(-1)


        print("Building website")
        for page in website.pages:
            page_links = links[page[PageAttributes.NAME]]

            for page_link in page_links:
                save_path = os.path.join(path,page_link)
                templates.generate_page(
                    page[PageAttributes.TEMPLATE],
                    save_path,
                    page_link,
                    target = website.get_target(page_link)
                )

        print("Built!")


    @staticmethod
    def template(path):
        """ Generates a template for a website """

        print("Building template")
        # create target folder if it does not exist
        if os.path.isdir(path):
            shutil.rmtree(path)
        os.mkdir(path)

        temp_name = "templates"
        asset_name = "assets"
        os.mkdir(os.path.join(path,temp_name))
        os.mkdir(os.path.join(path,asset_name))

        model_name = "model.json"
        model_dict = {"example":"hello world"}
        with open(os.path.join(path, model_name), 'w', encoding="utf-8") as model_out:
            json.dump(model_dict, model_out, indent=4)

        template_name = "template.html"
        with open(os.path.join(path, temp_name, template_name), 'w', encoding="utf-8") as temp_out:
            temp_out.write("<html>\n\t<head>\n\t\t<title>Example</title>\n\t</head>")
            temp_out.write("\n\t<body>\n\t\t<h1>{{ model['example'] }}</h1>")
            temp_out.write("\n\t</body>\n</html>")

        website = Website()
        website.name = "insert name"
        website.template_path = temp_name
        website.asset_paths = [asset_name]
        website.pages = [{
                'name':'example',
                'template':template_name,
                'filename':'example.html'
            }]

        with open(os.path.join(path,"website.json"), 'w', encoding="utf-8") as website_out:
            json.dump(website.to_dict(model_name), website_out, indent=4)

        print("Built")



"""
Responsible for managing the templating system and generating pages
"""

import datetime

import markdown
from jinja2 import FileSystemLoader
from jinja2 import Environment
from jinja2.exceptions import TemplateNotFound
from jinja2.filters import do_mark_safe


class TemplateEngine():
    """Provides a wrapper for the templating engine"""
    def __init__(self,template_path):
        self.env = Environment(
            loader = FileSystemLoader(template_path),
            autoescape = True
        )
        self.env.filters["date_fmt"] = TemplateEngine._date_filter
        self.env.filters["markdown_fmt"] = TemplateEngine._markdown_to_html5_filter

        self._template_path = template_path
        self._model = {}
        self._links = {}

    @staticmethod
    def _date_filter(
        date_in,
        date_format="%d %b %y",
        date_in_format="%Y-%m-%d"):
        """ Formats dates """
        if isinstance(date_in, str):
            date_in = datetime.datetime.strptime(date_in, date_in_format)

        return date_in.strftime(date_format)

    @staticmethod
    def _markdown_to_html5_filter(md_text):
        """ converts markdown strings to html5"""
        return do_mark_safe(
            markdown.markdown(md_text, output_format="html5")
            )

    @property
    def template_path(self):
        """ get template_path """
        return self._template_path

    @property
    def model(self):
        """ get model """
        return self._model

    @model.setter
    def model(self,value):
        """ sets model """
        self._model = value

    @property
    def links(self):
        """ get links """
        return self._links

    @links.setter
    def links(self,value):
        """ sets links """
        self._links = value

    def generate_page(self, template, save_path, page_link, target = None):
        """Uses templating engine to generate a new page and saves it to save_path"""
        template = self.env.get_template(template)

        page = template.render(
            model = self._model,
            links = self._links,
            page_link = page_link,
            target = target,
        )

        if save_path is not None:
            with open(save_path,'w', encoding="utf-8") as page_out:
                page_out.write(page)
            return page

        return page

    def has_template(self, template):
        """Checks to see if a template exists - returns True/False"""
        try:
            template = self.env.get_template(template)
            return template is not None
        except TemplateNotFound:
            return False



""" pytest module to hold tests for webstaterator/cli.py and webstaterator/__main__.py """

import os
import shutil
import argparse

import pytest


from webstaterator.webstaterator import Webstaterator
from webstaterator.cli import get_parser_options
from webstaterator.__main__ import *

EXPECTED_HELP_OUTPUT = """usage: webstaterator [-h] {validate,build,template} ...

A Python tool for generating static websites based on object models

optional arguments:
  -h, --help            show this help message and exit

Actions:
  Webstaterator actions

  {validate,build,template}
    validate            Validates the provided website description file
    build               Builds a website in the given folder based on the
                        provided website description file
    template            Generates a blank website description file
"""

def test_get_parser_returns_an_argument_parser():
    assert type(get_parser_options()) == argparse.ArgumentParser

def test_build_commands():
    pargs = get_parser_options().parse_args(
        ["build", "--website=test.json","--output=test/"]
    )

    assert pargs.action == "build"
    assert pargs.website == "test.json"
    assert pargs.output == "test/"

def test_validate_commands():
    pargs = get_parser_options().parse_args(
        ["validate", "--website=test.json"]
    )

    assert pargs.action == "validate"
    assert pargs.website == "test.json"

def test_template_commands():
    pargs = get_parser_options().parse_args(
        ["template", "--output=test/"]
    )

    assert pargs.action == "template"
    assert pargs.output == "test/"

def helper_cli_run_with_args(args, expected_output, capsys):
    run(args)
    out, err = capsys.readouterr()
    assert out == expected_output

def helper_cli_run_with_args_and_exit(args, expected_output, capsys):
    with pytest.raises(SystemExit):
        helper_cli_run_with_args(args, expected_output, capsys)

def test_cli_run_without_args(capsys):
    helper_cli_run_with_args_and_exit(
        [],
        EXPECTED_HELP_OUTPUT,
        capsys
    )

def test_cli_run_with_build(capsys,tmpdir):
    website_path = os.path.join(tmpdir, "test_website")
    ws_obj = Webstaterator()
    ws_obj.template(website_path)
    build_path = os.path.join(tmpdir,"build_output")
    website_json = os.path.join(website_path,"website.json")
    helper_cli_run_with_args(
        ["build", f"--output={build_path}", f"--website={website_json}"],
        """Building template
Built
Loading ...
copying assets
Building website
Built!
""",
        capsys
    )



"""
Webstaterator

A Python tool for generating static websites based on object models

Documentation: https://gitlab.com/Jon.Keatley.Folio/webstaterator
Gitlab: https://gitlab.com/Jon.Keatley.Folio/webstaterator
PyPi: https://pypi.org/project/webstaterator/

Created by Jon Keatley (http://jon-keatley.com)
Named by Sasha Siegel. It is her fault!

Copyright Jon Keatley 2021

"""



""" pytest module to hold tests for webstaterator/templateengine.py """

import os
import shutil
import datetime

import pytest

from webstaterator.templateengine import TemplateEngine

def make_template(path, file, template):
    temp_path = os.path.join(path,file)
    with open(temp_path,'w') as temp_out:
        temp_out.write(template)
    return temp_path

def make_list_template(path, file, name):
    with open(os.path.join(path,file),'w') as temp_out:
        temp_out.write("{%for item in "+name+" %}{{ item }}{% endfor %}")

def make_dict_template(path, file, name):
    with open(os.path.join(path,file),'w') as temp_out:
        temp_out.write(
            "{%for k,v in "+name+".items() %}{{ k }}={{ v }}---{% endfor %}"
        )

def make_date_filter_template(path, file, name, in_format = None):
    with open(os.path.join(path,file),'w') as temp_out:
        if in_format is None:
            temp_out.write("{{ model['"+name+"']|date_fmt }}")
        else:
            temp_out.write('{{ model["'+name+'"]|date_fmt("%d %b %y","'+ in_format +'")}}')

def make_markdown_filter_template(path, file, name):
    with open(os.path.join(path,file),'w') as temp_out:
            temp_out.write("{{ model['"+name+"']|markdown_fmt }}")


@pytest.fixture()
def get_test_template_engine():
    path = os.path.join(os.getcwd(),'test_templates')

    if(os.path.exists(path)):
        shutil.rmtree(path)

    os.mkdir(path)

    make_template(path, "basic.html", "test")
    make_dict_template(path, "model.html", "model")
    make_list_template(path, "links.html","links")
    make_dict_template(path, "target.html", "target")
    make_date_filter_template(path, "filter.html", "test_date")
    make_date_filter_template(path, "filter_cust.html", "test_date","%d-%m-%Y")
    make_markdown_filter_template(path, "filter_markdown.html","test_markdown")

    yield TemplateEngine(path)
    shutil.rmtree(path)

def test_create_website_instance(get_test_template_engine):
    assert isinstance(get_test_template_engine,TemplateEngine)

def test_properties(tmpdir):
    engine = TemplateEngine(tmpdir)
    model = {"result":"success"}
    engine.model = model
    assert engine.model == model
    links = ["this.html"]
    engine.links = links
    assert engine.links == links

def test_basic_template(get_test_template_engine):
    get_test_template_engine.model = {}
    get_test_template_engine.links = []
    result = get_test_template_engine.generate_page(
        template='basic.html',
        save_path = None,
        page_link = 'basic.html'
        )
    assert result == "test"

def test_links(get_test_template_engine):
    links = ['one.html','two.html']
    get_test_template_engine.model = {}
    get_test_template_engine.links = links
    result = get_test_template_engine.generate_page(
        template='links.html',
        save_path = None,
        page_link = "one.html"
        )
    assert result == "".join(links)

def test_basic_model(get_test_template_engine):

    model = {
        "a":1,
        "b":2,
        "c":3
    }

    model_flattened = ""
    for k,v in model.items():
        model_flattened += "{}={}---".format(k,v)

    get_test_template_engine.model = model
    get_test_template_engine.links = {}
    result = get_test_template_engine.generate_page(
        template='model.html',
        save_path = None,
        page_link = "model.html"
        )
    assert result == model_flattened

def test_page_link(get_test_template_engine):

    make_template(
        get_test_template_engine.template_path,
        "page_link.html",
        """{% for page,links in links.items() %}{% if page_link in links %}{{ page }}{% endif %}{% endfor %}"""
    )
    get_test_template_engine.model = {}
    get_test_template_engine.links = {
        "page1":["page1.html"],
        "page2":["page2.html"],
        "target":["target.html"]
    }
    result = get_test_template_engine.generate_page(
        template='page_link.html',
        target = None,
        save_path = None,
        page_link = "target.html"
        )
    assert result == "target"

def test_target(get_test_template_engine):
    target = {
        "ta":1,
        "tb":2,
        "tc":3
    }

    target_flattened = ""
    for k,v in target.items():
        target_flattened += "{}={}---".format(k,v)

    get_test_template_engine.model = {}
    get_test_template_engine.links = {}
    result = get_test_template_engine.generate_page(
        template='target.html',
        target = target,
        save_path = None,
        page_link = "target.html"
        )
    assert result == target_flattened

def test_saving_page(get_test_template_engine):
    page_link = "test_page.html"
    save_path = os.path.join(os.getcwd(),page_link)
    get_test_template_engine.model = {},
    get_test_template_engine.links = [],
    result = get_test_template_engine.generate_page(
        template='basic.html',
        save_path = save_path,
        page_link = page_link
        )

    assert result == "test"
    assert os.path.exists(save_path)

    with open(save_path,"r") as page_stream:
        assert page_stream.read() == "test"

    os.remove(save_path)

def test_has_template_success(get_test_template_engine):
    assert get_test_template_engine.has_template("model.html")

def test_has_template_fail(get_test_template_engine):
    assert not get_test_template_engine.has_template("no_template_found.html")

def test_date_filter_directly_with_date(get_test_template_engine):
    test_date = datetime.date(2021,1,1)
    test_date =  get_test_template_engine._date_filter(test_date)
    assert test_date == "01 Jan 21"

def test_date_filter_directly_with_datetime(get_test_template_engine):
    test_date = datetime.datetime(2021,2,2,1,0,0,0)
    test_date =  get_test_template_engine._date_filter(test_date)
    assert test_date == "02 Feb 21"

def test_date_filter_directly_with_str(get_test_template_engine):
    test_date = get_test_template_engine._date_filter("2021-03-03")
    assert test_date == "03 Mar 21"

def test_date_filter_in_template(get_test_template_engine):
    get_test_template_engine.model = {"test_date":"2021-04-04"}
    get_test_template_engine.links = []
    result = get_test_template_engine.generate_page(
        template="filter.html",
        save_path = None,
        page_link = "filter.html"
        )
    assert result == "04 Apr 21"

def test_date_filter_in_template_with_custom_input_format(get_test_template_engine):
    get_test_template_engine.model = {"test_date":"04-04-2021"}
    get_test_template_engine.links = []
    result = get_test_template_engine.generate_page(
        template="filter_cust.html",
        save_path = None,
        page_link = "filter_cust.html"
        )
    assert result == "04 Apr 21"

def test_markdown_filter_directly(get_test_template_engine):
    markdown = "# Header"
    html     = "<h1>Header</h1>"
    test_md = get_test_template_engine._markdown_to_html5_filter(markdown)
    assert test_md == html

def test_markdown_filter_in_teamplate(get_test_template_engine):
    markdown = "## Header"
    html     = "<h2>Header</h2>"
    get_test_template_engine.model = {"test_markdown":markdown}
    get_test_template_engine.links = []
    result = get_test_template_engine.generate_page(
    template="filter_markdown.html",
    save_path = None,
    page_link = "filter_markdown.html"
    )
    assert result == html
